# XLE in Docker

Maintainer: Jakob Moser <moser@cl.uni-heidelberg.de>

Docker image containing the [Xerox Linguistic Environment (XLE)][xle] to parse sentences using a Lexical Functional Grammar. 

Designed to be used in a "functional" style: You start a container from this image, pass a grammar and a sentence as input parameter and get the C- and F-structure in return. The container then terminates.

## :whale: Build the image

To run a container from this image, you first have to build it. To do so, clone the repository, change into it and then execute

```bash
sudo docker build -t xle .
```

This will create an image from the [`Dockerfile`](Dockerfile) and tag it as `xle`.

## :runner: Run a container

To parse a sentence using xle-docker, you need two things:

* A grammar, e.g. stored in the file `/home/mustermann/grammatik.lfg`
* A sentence, e.g. `in der Arena veranstalten die Ritter ein Turnier .`

You can then execute the following command:

```bash
sudo docker run \
     --mount type=bind,source=/home/mustermann/grammatik.lfg,target=/grammar.lfg \
     xle "in der Arena veranstalten die Ritter ein Turnier ."
```

This will take a few seconds. The parses will then be output to the command line using XLE's own XML export format. If you want to store the output in a file, append `> name_of_your_file.xml` at the end of the command.

If no parses are generated, the container will exit with exit code `42` and write the XLE shell output to stderr.

<details>
<summary>Advanced information</summary>

Note that the `xle` specified in the `docker run` command is the name of the tag you've given the image when building, _not_ the command to be executed (although under the hood the container will call a Linux command also called `xle`).
</details>

<details>
<summary>Alternative using a Base64-encoded JSON</summary>

If you don't want to use mounts, you can also create a JSON file that contains the attributes `sentence` and `grammar`, like this (shortened version, we just want to illustrate how to deal with newlines):

```json
{"sentence":"in der Arena veranstalten die Ritter ein Turnier", "grammar":"DEMO GERMAN CONFIG (1.0)\n    ROOTCAT IP.\n    LEXENTRIES (DEMO GERMAN).\n    RULES (DEMO GERMAN).\n    GOVERNABLERELATIONS SUBJ OBJ OBJ-TH OBL-TH.\n    SEMANTICFUNCTIONS ADJ.\n    NONDISTRIBUTIVES NUM PERS GEND.\n----\n\nDEMO GERMAN RULES (1.0)\n\nIP --> NP: (^ SUBJ) = !;\n       I'\n       (PUNCT).\n\nNP --> { Det N | NP PP: ! $ (^ ADJ); }.\n\nI' --> I VP.\n\n..."} 
```

Store the file as `data.json`. You can then run the container like this:

```bash
sudo docker container run xle $(cat data.json | base64 -w0)
```

The idea to implement it that way came from [Christian Heusel](https://christian.heusel.eu/), thank you very much!
</details>

## :abcd: A word about encoding

By default, this container lets XLE use its default encoding, whatever that is (not UTF-8). If you want to use UTF-8, you have to add the following line to your grammar file in the `CONFIG` section

```
CHARACTERENCODING utf-8.
```

**and** execute `set-character-encoding stdio utf-8` every time after starting XLE.

`xle-docker` will automatically use `utf-8` as character encoding for standard input if it detects the character encoding configuration line (more precisely, a line matching `^\s*CHARACTERENCODING\s+utf-8\s*\..*$` anywhere in the file).

## :page_with_curl: About

Here, a more detailed explanation of the relevant files is given. The used sources are marked with :books:.

### `Dockerfile`

The Dockerfile specifies the steps with which the image is built. As a first step, we install dependencies:

* `xvfb`: X server, because XLE won't start without one running
* `ca-certificates`: To verify the HTTPS connection to the external XLE repository

:books: [Running X applications headless using Xvfb][xvfb] <br />
:books: [Fake X server for testing?][xvfb-so]

After that, we configure the external repository "Zinnwerk" which provides us with a copy of the XLE application.

:books: [Instructions to connect to a third-party repository][3rd-repo-debwiki] <br />
:books: [How to add a third-party repo. and key in Debian?][3rd-repo-se] <br />
:books: [apt-key Is Deprecated. How To Add OpenPGP Repository Signing Keys Without It On Debian, Ubuntu, Linux Mint, Pop!_OS, Etc.][3rd-repo-lu]

As a last step, we configure the entry point. The entry point is a script that is started inside the container whenever someone creates runs a container.

### `entrypoint.sh`

This script is run inside the Docker container. It first creates an `xlerc` file containing instructions to create a parser for the specified grammar, parse the sentence, export the results and exit.

:books: [syntax.xle – syntaktischer zucker (I)][narnold-xle]

It then executes `xle` using the `xvfb-run` wrapper (which ensures `xle` can access an X server). `xle` executes the instructions contained in `xlerc`.

:books: [Ubuntu Manpage: xvfb-run - run specified X client or command in a virtual X server environment][xvfb-run]

As a final step, `cat` is used to output the export file to the standard output or to output the XLE log to standard error.

:books: [Answer to "echo that outputs to stderr" by Brandon Rhodes on Stack Overflow][stderr-so]

[xle]:https://ling.sprachwiss.uni-konstanz.de/pages/xle/
[xvfb]:https://blog.kagesenshi.org/2007/06/running-x-applications-headless-using.html
[xvfb-so]:https://stackoverflow.com/questions/14321636/fake-x-server-for-testing
[3rd-repo-debwiki]:https://wiki.debian.org/DebianRepository/UseThirdParty#Instructions_to_connect_to_a_third-party_repository
[3rd-repo-se]:https://unix.stackexchange.com/questions/332672/how-to-add-a-third-party-repo-and-key-in-debian
[3rd-repo-lu]:https://www.linuxuprising.com/2021/01/apt-key-is-deprecated-how-to-add.html
[narnold-xle]:https://www.cl.uni-heidelberg.de/~narnold/xle/posts/2021-05-25-syntax-xle-mad.html
[xvfb-run]:https://manpages.ubuntu.com/manpages/trusty/man1/xvfb-run.1.html
[stderr-so]:https://stackoverflow.com/a/11422223/
